(* Author: Vincent Penelle*)

(**
 * Solver for the game keen
 * 
 * This module uses Z3. It contains a function that generate a formula equivalent to the solved game given a representation of the game (containing its size and the starting clues), a function that calls the solver and returns the model (or the information there is none), and a function that, given a model and a game, displays the corresponding solved game.
 * 
 * Variables of the formula here will all be of type Arithmetic.Integer 
 * 
 * @see<https://www.chiark.greenend.org.uk/~sgtatham/puzzles/js/keen.html> Link to the game. 
 *)

(**
 * This type is used to represent a cell of the gameboard, and is given by two integers: [(i,j)] represents the cell at row i and column j.
 *)
type cell = int * int

(**
 * This type represents a zone of the boardgame. It is composed of a constructor representing the operation (Add for addition, Sub for substraction, Mul for multiplication and Div for division. Add and Mul contain a list of cells as zones labelled by them can contain an arbitrary number of cells, while Sub and Div contain a pair of cells as zones labelled by them can contain only two cells. 
 *)
type op = Add of cell list | Sub of cell * cell | Mul of cell list | Div of cell * cell

(** 
 * This type represents a clue. It contains an op stating what is the zone and operation concerned, and an integer representing the result.
 *)
type clue = op * int

(** 
 * This type represents a instance of an unruly game. [(size,clues)] represents the game whose gameboard has [size] rows and columns, and whose clues are the ones present in the list [clues]. 
 *)
type game = int * clue list

(**
 * A context to be used by Z3.
 *)
val ctx : Z3.context


(**
 * [formula_game game] gives a formula equivalent to the game passed in argument. Namely, it specifies that each row (resp. column) have their cells labelled with different numbers, that all their cells are between 1 and the size of the game and that the clues are respected. 
 * The variable representing the cell (i,j) will be represented by the variable named with the string ["i,j"]. E.g., the cell (4,5) is represented with ["4,5"]
 *)
val game_formula : game -> Z3.Expr.expr

(**
 * Calls the solver over the formula given in argument and return the resulting model (if it exists). 
 * The return value should be None if there is no solution, or Some model, if the formula is satisfiable, and model is the resulting model. 
 *)
val call_solver : Z3.Expr.expr -> Z3.Model.model option

(** 
 * Print the game board corresponding to the model given in argument, with the size of the grid is given by the game specified in argument. 
 * [print_game game model] displays a grid of size [size] × [size], where [game = (size,clues)]. 
 *)
val print_game : game -> Z3.Model.model -> unit
