(* Author: Vincent Penelle*)

(**
 * Solver for the game unruly.
 * 
 * This module uses Z3. It contains a function that generate a formula equivalent to the solved game given a representation of the game (containing its size and the starting clues), a function that calls the solver and returns the model (or the information there is none), and a function that, given a model and a game, displays the corresponding solved game.
 * 
 * Variables of the formula here will all be of type Boolean, with, e.g., true representing a white cell, and false a black cell (but do the contrary if you prefer. It is not the only solution, but the solver will run smoother with only booleans. 
 *
 * @see<https://www.chiark.greenend.org.uk/~sgtatham/puzzles/js/unruly.html> Link to the game.
 *)


(**
 * This type is used to represent a clue in the game. It simply is a cell of the gameboard, and is given by two integers: [(i,j)] represents the cell at row i and column j.
 *)
type clue = int * int

(** 
 * This type represents a instance of an unruly game. [(n,m,positive_clues,negative_clues)] represents the game whose gameboard has n rows and m columns, whose cells in positive_clues are white and cells in negative_clues are black. 
 *)
type game = int * int * clue list * clue list

(**
 * A context to be used by Z3.
 *)
val ctx : Z3.context

(**
 * [formula_game game] gives a formula equivalent to the game passed in argument. Namely, it specifies that each row and column have the same number of black and white cells, that no three consecutive cells have the same colors, and that the clues are respected. 
 * The variable representing the cell (i,j) will be represented by the variable named with the string ["i,j"]. E.g., the cell (4,5) is represented with ["4,5"]
 *)
val game_formula : game -> Z3.Expr.expr

(**
 * Calls the solver over the formula given in argument and return the resulting model (if it exists). 
 * The return value should be None if there is no solution, or Some model, if the formula is satisfiable, and model is the resulting model. 
 *)
val call_solver : Z3.Expr.expr -> Z3.Model.model option

(** 
 * Print the game board corresponding to the model given in argument, with the size of the grid is given by the game specified in argument. 
 * [print_game game model] displays a grid of size row × col, where [game = (row,col,p,n)]. 
 *)
val print_game : game -> Z3.Model.model -> unit